=== modified file '.bzrignore'
--- .bzrignore	2008-08-22 02:18:27 +0000
+++ .bzrignore	2008-09-03 06:31:11 +0000
@@ -41,6 +41,7 @@
 bzrlib/_btree_serializer_c.c
 bzrlib/_dirstate_helpers_c.c
 bzrlib/_knit_load_data_c.c
+bzrlib/_readdir_pyx.c
 bzrlib/_walkdirs_win32.c
 doc/en/release-notes/NEWS.txt
 doc/en/developer-guide/HACKING.txt

=== added file 'bzrlib/_readdir_py.py'
--- bzrlib/_readdir_py.py	1970-01-01 00:00:00 +0000
+++ bzrlib/_readdir_py.py	2008-09-02 04:45:47 +0000
@@ -0,0 +1,33 @@
+# Copyright (C) 2006, 2008 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Python implementation of readdir interface."""
+
+
+import os
+
+
+def read_dir(path):
+    """Like os.listdir, this reads the contents of a directory.
+
+    There is a C module which is recommended which will return
+    a sort key in the first element of the tuple to allow slightly
+    more efficient behaviour on the operating systems part.
+
+    :param path: the directory to list.
+    :return: a list of (None, basename) tuples.
+    """
+    return [(None, name) for name in os.listdir(path)]

=== added file 'bzrlib/_readdir_pyx.pyx'
--- bzrlib/_readdir_pyx.pyx	1970-01-01 00:00:00 +0000
+++ bzrlib/_readdir_pyx.pyx	2008-09-02 04:45:47 +0000
@@ -0,0 +1,117 @@
+# Copyright (C) 2006, 2008 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Wrapper for readdir which returns files ordered by inode."""
+
+
+import os
+import sys
+
+
+# the opaque C library DIR type.
+cdef extern from 'errno.h':
+    int ENOENT
+    int ENOTDIR
+    int EAGAIN
+    int errno
+    char *strerror(int errno)
+
+cdef extern from 'sys/types.h':
+    ctypedef long ssize_t
+    ctypedef unsigned long size_t
+
+cdef extern from 'dirent.h':
+    int DT_UNKNOWN
+    int DT_REG
+    int DT_DIR
+    int DT_FIFO
+    int DT_SOCK
+    int DT_CHR
+    int DT_BLK
+    ctypedef struct dirent:
+        char d_name[256]
+        # this will fail to compile if d_type is not defined.
+        # if this module fails to compile, use the .py version.
+        unsigned char d_type
+        int d_ino
+    ctypedef struct DIR
+    # should be DIR *, pyrex barfs.
+    DIR * opendir(char * name)
+    int closedir(DIR * dir)
+    dirent *readdir(DIR *dir)
+
+_directory = 'directory'
+_chardev = 'chardev'
+_block = 'block'
+_file = 'file'
+_fifo = 'fifo'
+_symlink = 'symlink'
+_socket = 'socket'
+_unknown = 'unknown'
+
+dot = ord('.')
+
+# add a typedef struct dirent dirent to workaround pyrex
+cdef extern from 'readdir.h':
+    pass
+
+def read_dir(path):
+    """Like os.listdir, this reads the contents of a directory.
+
+    :param path: the directory to list.
+    :return: a list of (sort_key, basename) tuples.
+    """
+    cdef DIR *the_dir
+    # currently this needs a fixup - the C code says 'dirent' but should say
+    # 'struct dirent'
+    cdef dirent * entry
+    cdef dirent sentinel
+    cdef char *name
+    the_dir = opendir(path)
+    if NULL == the_dir:
+        raise OSError(errno, strerror(errno))
+    result = []
+    try:
+        entry = &sentinel
+        while entry != NULL:
+            entry = readdir(the_dir)
+            if entry == NULL:
+                if errno == EAGAIN:
+                    # try again
+                    continue
+                elif errno != ENOTDIR and errno != ENOENT and errno != 0:
+                    # We see ENOTDIR at the end of a normal directory.
+                    # As ENOTDIR for read_dir(file) is triggered on opendir,
+                    # we consider ENOTDIR to be 'no error'.
+                    # ENOENT is listed as 'invalid position in the dir stream' for
+                    # readdir. We swallow this for now and just keep reading.
+                    raise OSError(errno, strerror(errno))
+                else:
+                    # done
+                    continue
+            name = entry.d_name
+            if not (name[0] == dot and (
+                (name[1] == 0) or 
+                (name[1] == dot and name[2] == 0))
+                ):
+                result.append((entry.d_ino, entry.d_name))
+    finally:
+        if -1 == closedir(the_dir):
+            raise OSError(errno, strerror(errno))
+    return result
+
+
+# vim: tw=79 ai expandtab sw=4 sts=4

=== modified file 'bzrlib/benchmarks/bench_osutils.py'
--- bzrlib/benchmarks/bench_osutils.py	2007-04-09 22:11:22 +0000
+++ bzrlib/benchmarks/bench_osutils.py	2008-08-20 02:07:36 +0000
@@ -29,6 +29,7 @@
         self.make_kernel_like_tree(link_working=True)
         # on roberts machine: this originally took:  157ms/4177ms
         # plain os.walk takes 213ms on this tree
+        # with the pyrex readdir module:              77ms/5423ms
         def dowalk():
             for dirblock in osutils.walkdirs('.'):
                 if dirblock[0][1] == '.bzr':

=== modified file 'bzrlib/osutils.py'
--- bzrlib/osutils.py	2008-09-02 05:28:37 +0000
+++ bzrlib/osutils.py	2008-09-03 06:31:11 +0000
@@ -1278,7 +1278,8 @@
     """
     _lstat = os.lstat
     _directory = _directory_kind
-    _listdir = os.listdir
+    # Use C accelerated directory listing.
+    _listdir = _read_dir
     _kind_from_mode = _formats.get
 
     # 0 - relpath, 1- basename, 2- kind, 3- stat, 4-toppath
@@ -1294,11 +1295,13 @@
 
         dirblock = []
         append = dirblock.append
-        for name in sorted(_listdir(top)):
+        # read_dir supplies in should-stat order.
+        for _, name in sorted(_listdir(top)):
             abspath = top_slash + name
             statvalue = _lstat(abspath)
             kind = _kind_from_mode(statvalue.st_mode & 0170000, 'unknown')
             append((relprefix + name, name, kind, statvalue, abspath))
+        dirblock.sort()
         yield (relroot, top), dirblock
 
         # push the user specified dirs from dirblock
@@ -1543,3 +1546,9 @@
         base = abspath(pathjoin(base, '..', '..'))
     filename = pathjoin(base, resource_relpath)
     return open(filename, 'rU').read()
+
+
+try:
+    from bzrlib._readdir_pyx import read_dir as _read_dir
+except ImportError:
+    from bzrlib._readdir_py import read_dir as _read_dir

=== added file 'bzrlib/readdir.h'
--- bzrlib/readdir.h	1970-01-01 00:00:00 +0000
+++ bzrlib/readdir.h	2008-09-02 04:45:47 +0000
@@ -0,0 +1,22 @@
+/*
+ *  Bazaar -- distributed version control
+ *
+ * Copyright (C) 2006 by Canonical Ltd
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
+
+/* Adjust C api to workaround pyrex output bug/limitation */
+typedef struct dirent dirent;

=== modified file 'bzrlib/tests/test_osutils.py'
--- bzrlib/tests/test_osutils.py	2008-08-15 05:35:17 +0000
+++ bzrlib/tests/test_osutils.py	2008-09-03 06:31:11 +0000
@@ -40,11 +40,15 @@
         pump_string_file,
         )
 from bzrlib.tests import (
+        adapt_tests,
+        Feature,
         probe_unicode_in_user_encoding,
+        split_suite_by_re,
         StringIOWrapper,
         SymlinkFeature,
         TestCase,
         TestCaseInTempDir,
+        TestScenarioApplier,
         TestSkipped,
         )
 from bzrlib.tests.file_utils import (
@@ -53,6 +57,35 @@
 from bzrlib.tests.test__walkdirs_win32 import WalkdirsWin32Feature
 
 
+def load_tests(standard_tests, module, loader):
+    """Parameterize readdir tests."""
+    to_adapt, result = split_suite_by_re(standard_tests, "readdir")
+    adapter = TestScenarioApplier()
+    from bzrlib import _readdir_py
+    adapter.scenarios = [('python', {'read_dir': _readdir_py.read_dir})]
+    if ReadDirFeature.available():
+        adapter.scenarios.append(('pyrex',
+            {'read_dir': ReadDirFeature.read_dir}))
+    adapt_tests(to_adapt, adapter, result)
+    return result
+
+
+class _ReadDirFeature(Feature):
+
+    def _probe(self):
+        try:
+            from bzrlib import _readdir_pyx
+            self.read_dir = _readdir_pyx.read_dir
+            return True
+        except ImportError:
+            return False
+
+    def feature_name(self):
+        return 'bzrlib._btree_serializer_c'
+
+ReadDirFeature = _ReadDirFeature()
+
+
 class TestOSUtils(TestCaseInTempDir):
 
     def test_contains_whitespace(self):
@@ -746,6 +779,38 @@
 
 class TestWalkDirs(TestCaseInTempDir):
 
+    def test_readdir(self):
+        tree = [
+            '.bzr/',
+            '0file',
+            '1dir/',
+            '1dir/0file',
+            '1dir/1dir/',
+            '2file'
+            ]
+        self.build_tree(tree)
+        expected_names = ['.bzr', '0file', '1dir', '2file']
+        # read_dir returns pairs, which form a table with either None in all
+        # the first columns, or a sort key to get best on-disk-read order, 
+        # and the disk path name in utf-8 encoding in the second column.
+        read_result = self.read_dir('.')
+        # The second column is always the names, and every name except "." and
+        # ".." should be present.
+        names = sorted([row[1] for row in read_result])
+        self.assertEqual(expected_names, names)
+        expected_sort_key = None
+        if read_result[0][0] is None:
+            # No sort key returned - all keys must None
+            operator = self.assertEqual
+        else:
+            # A sort key in the first row implies sort keys in the other rows.
+            operator = self.assertNotEqual
+        for row in read_result:
+            operator(None, row[0])
+
+    def test_compiled_extension_exists(self):
+        self.requireFeature(ReadDirFeature)
+        
     def test_walkdirs(self):
         tree = [
             '.bzr',

=== modified file 'setup.py'
--- setup.py	2008-08-22 02:18:27 +0000
+++ setup.py	2008-09-03 06:31:11 +0000
@@ -228,6 +228,7 @@
 add_pyrex_extension('bzrlib._btree_serializer_c')
 add_pyrex_extension('bzrlib._dirstate_helpers_c')
 add_pyrex_extension('bzrlib._knit_load_data_c')
+add_pyrex_extension('bzrlib._readdir_pyx')
 if sys.platform == 'win32':
     # pyrex uses the macro WIN32 to detect the platform, even though it should
     # be using something like _WIN32 or MS_WINDOWS, oh well, we can give it the

